import asyncio
import datetime
import json
import logging
import os
import re
import ssl
import time

import aiofiles
from aiohttp import web
from aiohttp.web_exceptions import HTTPUnauthorized

class LogviewServer:
  def __init__(self, config, http_port, debug):
    self.config = config
    self.http_port = config.get("server_http_port", http_port)
    self.ssl_key = config.get("server_ssl_key")
    self.ssl_cert = config.get("server_ssl_cert")
    self.access_logs = config.get("access_logs", debug)
    self.log_expiration_seconds = int(self.config.get("log_expiration_seconds", -1))
    self.log = logging.getLogger(__name__)

    self.log.info("http_port [%d], debug [%s], config [%s]", self.http_port, debug,
                  re.sub(r"(.*?'auth_token': ').*? ", r"\1**********', ", str(self.config)))

  def run(self):
    ssl_context = None
    if (self.ssl_key is not None and self.ssl_cert is not None):
      ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
      ssl_context.load_cert_chain(self.ssl_cert, self.ssl_key)

    loop = asyncio.get_event_loop()

    http_handler = LogviewHttpServer(self).new_handler()
    http_server = loop.run_until_complete(loop.create_server(http_handler, '0.0.0.0', self.http_port, ssl=ssl_context))

    try:
      loop.run_forever()
    finally:
      http_server.close()
      loop.run_until_complete(http_server.wait_closed())
      loop.close()

class LogviewHttpServer:
  def __init__(self, parent):
    self.parent = parent
    self.config = parent.config
    self.app = web.Application()
    self.app.router.add_route('GET', '/', self.get_root)
    self.app.router.add_route('GET', '/logs', self.get_log_targets)
    self.app.router.add_route('GET', '/logs/', self.get_log_targets)
    self.app.router.add_route('GET', r'/logs/{name:[^{}]+}', self.get_log_file)
    self.log = logging.getLogger(__name__)

  def new_handler(self):
    return self.app.make_handler() if self.parent.access_logs else self.app.make_handler(access_log=None)

  def check_auth(self, request):
    auth_token = self.config.get("auth_token")
    if not auth_token:
      return

    request_auth = request.headers.get(self.config.get("auth_header", "Imply-Token"))
    if not request_auth or request_auth != auth_token:
      raise HTTPUnauthorized

  async def get_root(self, request):
    return web.Response(status=200, body=datetime.datetime.now(datetime.timezone.utc).isoformat())

  async def get_log_targets(self, request):
    self.check_auth(request)
    now = time.time()

    readable_files = {k for (k, v) in self.config.get("targets", {}).items() if
                      'path' in v and os.access(v["path"], os.R_OK) and (
                          self.parent.log_expiration_seconds < 0
                          or now - self.parent.log_expiration_seconds < os.path.getmtime(v["path"]))}
    detailed = []
    for key in readable_files:
      m_time = os.path.getmtime(self.config["targets"][key]["path"])

      detailed.append({"name": key, "modifiedTime": int(round(m_time * 1000)), **self.config["targets"][key]})

    resp = {
      'logs': list(readable_files),
      'detailed': detailed
    }

    return web.Response(status=200, body=json.dumps(resp), content_type="application/json")

  async def get_log_file(self, request):
    self.check_auth(request)
    buf_sz = int(self.config.get("buffer_size", 65536))

    log_name = request.match_info['name']
    readable_files = {k for (k, v) in self.config.get("targets", {}).items() if
                      'path' in v and os.access(v["path"], os.R_OK)}

    if log_name not in readable_files:
      return web.Response(status=404, body=json.dumps({"error": "[{}] is not a valid log name".format(log_name)}),
                          content_type="application/json")

    try:
      offset = int(request.rel_url.query.get("offset", 0))
    except ValueError:
      return web.Response(status=400, body=json.dumps({"error": "[offset] must be a number"}),
                          content_type="application/json")

    resp = web.StreamResponse(status=200, headers={'Content-Type': 'text/plain'})
    await resp.prepare(request)

    async with aiofiles.open(self.config["targets"][log_name]["path"], 'rb') as fh:
      if offset < 0:
        await fh.seek(0, os.SEEK_END)
        offset = max(0, await fh.tell() + offset)

      await fh.seek(offset)

      # Unless we're at the beginning, discard the first line which will probably be truncated
      if self.config.get("seek_to_newline", True) and offset != 0:
        data = await fh.read(1)
        while data and data != b'\n':
          data = await fh.read(1)

      ba = bytearray(buf_sz)
      len = await fh.readinto(ba)
      while len:
        await resp.write(ba if len == buf_sz else ba[0:len])
        len = await fh.readinto(ba)

      await resp.write_eof()
    return resp
